#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# ISIS-DM: the ISIS Data Model API
#
# Copyright (C) 2010 BIREME/PAHO/WHO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .ordered import OrderedProperty, OrderedModel
from .subfield import CompositeString
import json
import colander
import deform

class Document(OrderedModel):

    def __init__(self, **kwargs):
        super(Document, self).__init__(**kwargs)
        self.validate()

    def validate(self):
        for prop in self:
            descriptor = self.__class__.__getattribute__(self.__class__, prop)
            descriptor.validate(self, getattr(self, prop, None))

    @classmethod
    def get_schema(cls):
        schema = colander.SchemaNode(colander.Mapping())
        exclude_fields = ()

        if hasattr(cls, 'Meta'):
            if hasattr(cls.Meta, 'hide'):
                if not isinstance(cls.Meta.hide, tuple):
                    raise TypeError('hide value must be tuple')
                exclude_fields = cls.Meta.hide

        for prop in cls:
            if prop in exclude_fields:
                continue
            descriptor = cls.__getattribute__(cls, prop)
            colander_definition = descriptor._colander_schema(cls, getattr(cls, prop, None))
            schema.add(colander_definition)

        return schema

    def to_python(self):
        '''
        generate a python representation for Document type classes
        '''
        properties = {}
        for prop in self:
            descriptor = self.__class__.__getattribute__(self.__class__, prop)
            properties[prop] = descriptor._pystruct(self, getattr(self, prop, None))
        if not 'TYPE' in properties:
            properties['TYPE'] = self.TYPE

        return properties

    @classmethod
    def from_python(cls, pystruct):
        if 'TYPE' in pystruct and cls.__name__ != pystruct.pop('TYPE'):
            raise TypeError()

        isisdm_pystruct = dict((str(k), tuple(v) if isinstance(v, list) else v)
            for k, v in pystruct.items())

        return cls(**isisdm_pystruct)

class Invalid(Exception):
    ''' TODO: study colander.Invalid exception '''
    def __init__(self, message):
        super(Exception, self).__init__(message)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.message)


class CheckedProperty(OrderedProperty):

    def __init__(self, required=False, validator=None, choices=None):
        super(CheckedProperty, self).__init__()
        self.required = required
        self.validator = validator
        self.choices = choices if choices else []

    def __set__(self, instance, value):
        if self.validator:
            self.validator(self, value)
        super(CheckedProperty, self).__set__(instance, value)

    def validate(self, instance, value):
        if self.missing(instance):
            raise TypeError('required %r property missing' % self.name)
        if value is not None and self.validator:
            self.validator(self, value)

    def missing(self, instance):
        # a property is missing if it is required and
        # does not exist or is set to None
        # if it is not required, it is never missing
        return self.required and getattr(instance, self.name, None) is None


class TextProperty(CheckedProperty):

    def __set__(self, instance, value):
        if not isinstance(value, basestring):
            raise TypeError('%r value must be unicode or str instance' % self.name)
        value = unicode(value)
        if self.required and len(value.rstrip()) == 0:
            raise Invalid('%r value cannot be empty' % self.name)
        super(TextProperty, self).__set__(instance, value)

    def _pystruct(self, instance, value):
        '''
        python representation for this property
        '''
        return value

    def _colander_schema(self, instance, value):
        return colander.SchemaNode(colander.String(), name=self.name)

class BooleanProperty(CheckedProperty):

    def __set__(self, instance, value):
        if not isinstance(value, bool):
            raise TypeError('%r must be bool' % self.name)
        super(BooleanProperty, self).__set__(instance, value)

    def _pystruct(self, instance, value):
        '''
        python representation for this property
        '''
        return value

    def _colander_schema(self, instance, value):
        return colander.SchemaNode(colander.Boolean(), name=self.name)

class FileProperty(CheckedProperty):

    def __set__(self, instance, value):
        if not isinstance(value, dict):
            raise TypeError('%r must be a dictionary' % self.name)
        
        if 'fp' not in value:
            raise TypeError('fp value must exists')
        
        if 'filename' not in value:
            try:
                value['filename'] = value['fp'].name
            except AttributeError:
                raise TypeError('%r must be a file' % self.name)


        super(FileProperty, self).__set__(instance, value)

    def _pystruct(self, instance, value):
        '''
        python representation for this property
        '''
        value['fp'] = None        
        return value

    def _colander_schema(self, instance, value):
        class MemoryTmpStore(dict):
            def preview_url(self, name):
                return None
        tmpstore = MemoryTmpStore()
        return colander.SchemaNode(deform.FileData(), 
                                   widget=deform.widget.FileUploadWidget(tmpstore),
                                   name=self.name)

class MultiTextProperty(CheckedProperty):

    def __set__(self, instance, value):
        if not isinstance(value, tuple):
            raise TypeError('MultiText value must be tuple')
        super(MultiTextProperty, self).__set__(instance, value)

    def _pystruct(self, instance, value):
        '''
        python representation for this property
        '''
        return value

    def _colander_schema(self, instance, value):
        schema = colander.SchemaNode(colander.Sequence(), name=self.name)
        schema.add(colander.SchemaNode(colander.String(), name=self.name))

        return schema


class CompositeTextProperty(CheckedProperty):

    def __init__(self, subkeys=None, **kwargs):
        super(CompositeTextProperty, self).__init__(**kwargs)
        self.subkeys = subkeys

    def __set__(self, instance, value):
        if not isinstance(value, basestring):
            raise TypeError('%r value must be unicode or str instance' % self.name)

        composite_text = CompositeString(value, self.subkeys)
        super(CompositeTextProperty, self).__set__(instance, composite_text)

    def _pystruct(self, instance, value):
        '''
        python representation for this property
        '''
        return value.items()

    def _colander_schema(self, instance, value):
        subfield = colander.SchemaNode(colander.Tuple())
        subfield.add(colander.SchemaNode(colander.String(), name='subkey'))
        subfield.add(colander.SchemaNode(colander.String(), name='value'))

        return colander.SchemaNode(colander.Sequence(),
                                   subfield,
                                   name=self.name)

class MultiCompositeTextProperty(CheckedProperty):

    def __init__(self, subkeys=None, **kwargs):
        super(MultiCompositeTextProperty, self).__init__(**kwargs)
        self.subkeys = subkeys

    def __set__(self, instance, value):
        if not isinstance(value, tuple):
            raise TypeError('MultiCompositeText value must be tuple')

        composite_texts = tuple(CompositeString(raw_composite_text, self.subkeys) for raw_composite_text in value)
        super(MultiCompositeTextProperty, self).__set__(instance, composite_texts)

    def _pystruct(self, instance, value):
        '''
        python representation for this property
        '''
        return tuple(composite_text.items() for composite_text in value)

    def _colander_schema(self, instance, value):
        schema = colander.SchemaNode(colander.Tuple())
        for subkey in self.subkeys:
            schema.add(colander.SchemaNode(colander.String(), name=subkey))

        return colander.SchemaNode(colander.Sequence(),
                                   schema,
                                   name=self.name)

class ReferenceProperty(CheckedProperty):

    def __set__(self, instance, value):
        if not isinstance(value, basestring):
            raise TypeError('Reference value must be unicode or str instance')
        value = unicode(value)
        if self.required and len(value.rstrip()) == 0:
            raise Invalid('%r value cannot be empty' % self.name)
        super(ReferenceProperty, self).__set__(instance, value)

    def _pystruct(self, instance, value):
        '''
        python representation for this property
        '''
        return value

    def _colander_schema(self, instance, value):
        return [colander.String()]

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
--------------------------------
Persisting a document in CouchDB
--------------------------------
    >>> def text_validator(node, value):
    ...     if value.startswith('Banana'):
    ...         raise Exception("You can't start a text with 'Banana'")
    ...

    >>> def colon_validator(node, value):
    ...     for author in value:
    ...         if ',' not in author:
    ...             raise Exception("Authors name must be in 'LastName, FirstName' format")

    >>> class Book(CouchdbDocument):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = MultiTextProperty(required=False, validator=colon_validator)
    ...     pages = TextProperty()

    >>> class BookWithAttachment(CouchdbDocument):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = MultiCompositeTextProperty(required=False, subkeys='fl')
    ...     cover = FileProperty()
    ...

Instantiating a Book object::

    >>> book1 = Book(title='Godel, Escher, Bach',
    ...               authors=(u'Hofstadter, Douglas',),
    ...               pages='777',)
    ...

    >>> len(book1._id)
    5

    >>> for key in sorted(book1.to_python()): print key, book1.to_python()[key] #doctest: +ELLIPSIS
    TYPE Book
    _id ...
    authors (u'Hofstadter, Douglas',)
    pages 777
    title Godel, Escher, Bach

    >>> book1_id = book1.save(db)

    >>> book2 = Book.get(db, book1_id)
    >>> book2.title
    u'Godel, Escher, Bach'

    >>> book3 = BookWithAttachment(title='Other Ninar songs',
    ...               authors=(u'^lHeineman^fGeorge T.',
    ...                        u'^lPollice^fGary',
    ...                        u'^lSelkov^fStanley'),
    ...               cover={'fp':open('test_mapper.py')})
    ...
    >>> book3_id = book3.save(db)

    >>> book_schema = BookWithAttachment.get_schema()
    >>> ' '.join('%s:%s' % (c.name, type(c.typ).__name__) for c in book_schema.children)
    'title:String authors:Sequence cover:FileData _rev:String _id:String'

"""
from isis.model import CouchdbDocument
from isis.model import TextProperty, MultiTextProperty
from isis.model import CompositeTextProperty, MultiCompositeTextProperty, ReferenceProperty, FileProperty
import couchdbkit


def test():
    import doctest
    doctest.testmod()

if __name__=='__main__':
    conn = couchdbkit.Server('http://127.0.0.1:5984')
    db = conn.get_or_create_db('isismodel_api')
    test()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# ISIS-DM: the ISIS Data Model API
#
# Copyright (C) 2010 BIREME/PAHO/WHO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re


MAIN_SUBFIELD_KEY = '_'
SUBFIELD_MARKER_RE = re.compile(r'\^([a-z0-9])', re.IGNORECASE)
DEFAULT_ENCODING = u'utf-8'

def expand(content, subkeys=None):
    ''' Parse a field into an association list of keys and subfields

        >>> expand('zero^1one^2two^3three')
        [('_', 'zero'), ('1', 'one'), ('2', 'two'), ('3', 'three')]

    '''
    if subkeys is None:
        regex = SUBFIELD_MARKER_RE
    elif subkeys == '':
        return [(MAIN_SUBFIELD_KEY, content)]
    else:
        regex = re.compile(r'\^(['+subkeys+'])', re.IGNORECASE)
    content = content.replace('^^', '^^ ')
    parts = []
    start = 0
    key = MAIN_SUBFIELD_KEY
    while True:
        found = regex.search(content, start)
        if found is None: break
        parts.append((key, content[start:found.start()].rstrip()))
        key = found.group(1).lower()
        start = found.end()
    parts.append((key, content[start:].rstrip()))
    return parts


class CompositeString(object):
    ''' Represent an Isis field, with subfields, using
        Python native datastructures

        >>> author = CompositeString('John Tenniel^xillustrator',
        ...                          subkeys='x')
        >>> unicode(author)
        u'John Tenniel^xillustrator'

    '''

    def __init__(self, isis_raw, subkeys=None, encoding=DEFAULT_ENCODING):
        if not isinstance(isis_raw, basestring):
            raise TypeError('%r value must be unicode or str instance' % isis_raw)

        self.__isis_raw = isis_raw.decode(encoding)
        self.__expanded = expand(self.__isis_raw, subkeys)

    def __getitem__(self, key):
        for subfield in self.__expanded:
            if subfield[0] == key:
                return subfield[1]
        else:
            raise KeyError(key)

    def __iter__(self):
        return (subfield[0] for subfield in self.__expanded)

    def items(self):
        return self.__expanded

    def __unicode__(self):
        return self.__isis_raw

    def __str__(self):
        return str(self.__isis_raw)



def test():
    import doctest
    doctest.testmod()

if __name__=='__main__':
    test()



# package
from .mapper import Document
from .mapper import TextProperty, MultiTextProperty
from .mapper import CompositeTextProperty, MultiCompositeTextProperty
from .mapper import ReferenceProperty, FileProperty, BooleanProperty
from .couchdb import CouchdbDocument


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# ISIS-DM: the ISIS Data Model API
#
# Copyright (C) 2010 BIREME/PAHO/WHO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from ..utils import base28
from .mapper import Document, TextProperty, FileProperty
import uuid
import couchdbkit
import time
import colander
import deform

class CouchdbDocument(Document):

    def __init__(self, **kwargs):
        super(CouchdbDocument, self).__init__(**kwargs)
        if '_id' not in kwargs:
            self._id = base28.genbase(5)

    def __clean_before_save(self, doc):
        '''
        removes attributes with None values
        '''
        doc = dict((str(k), v) for k, v in doc.items() if v is not None)

        return doc

    def to_python(self):
        pstruct = super(CouchdbDocument, self).to_python()
        pstruct['_id'] = self._id
        
        try:
            pstruct['_rev'] = self._rev
        except AttributeError:
            pass
        
        return pstruct

    @classmethod
    def from_python(cls, pystruct):
        if pystruct['_id'] == 'None':
            pystruct['_id'] = None
        if pystruct['_rev'] == 'None':
            pystruct['_rev'] = None
        
        return super(CouchdbDocument, cls).from_python(pystruct)

    def save(self, db):

        doc = self.to_python()

        doc = self.__clean_before_save(doc)

        while True:
            try:
                db.save_doc(doc)
                for key in self.__class__:
                    prop = self.__class__.__getattribute__(self.__class__,key)
                    if isinstance(prop, FileProperty):                        
                        file_dict = getattr(self,key)
                        db.put_attachment(doc, file_dict['fp'], getattr(self, key)['filename'])                        
                break
            except couchdbkit.ResourceConflict:
                time.sleep(0.5)
                doc['_id'] = base28.genbase(5)

        return doc['_id']
    
    @classmethod
    def get(cls, db, doc_id):
        doc = db.get(doc_id)
        return cls.from_python(doc)

    @classmethod
    def get_schema(cls):
        schema = super(CouchdbDocument, cls).get_schema()
        rev_definition = colander.SchemaNode(colander.String(),
                                                  widget = deform.widget.HiddenWidget(),
                                                  default=None,
                                                  name='_rev')
        id_definition = colander.SchemaNode(colander.String(),
                                                  widget = deform.widget.HiddenWidget(),
                                                  default=None,
                                                  name='_id')                                        
        schema.add(rev_definition)
        schema.add(id_definition)

        return schema


#!/usr/bin/env python
# coding: utf-8

import string
from random import randrange

BASE36 = string.digits+string.ascii_lowercase
# sem vogais, para nao formar palavras em portugues
# sem 1l0, para evitar confusões na leitura
BASE28 = ''.join(d for d in BASE36 if d not in '1l0aeiou') 

def reprbase(n, digitos=BASE28):
    ''' devolve a representação do valor `n` usando `digitos` '''
    base = len(digitos)
    s = []
    while n:
        n, d = divmod(n, base)
        s.insert(0, digitos[d])
    if s:    
        return ''.join(s)
    else:
        return digitos[0]
    
def calcbase(s, digitos=BASE28):
    ''' devolve o valor numérico de `s` na base representada pelos dígitos '''
    return sum(digitos.index(dig)*len(digitos)**pot
               for pot, dig in enumerate(reversed(s)))

def genbase(tamanho, digitos=BASE28):
    return reprbase(randrange(len(digitos)**tamanho), digitos).rjust(tamanho,digitos[0])

if __name__=='__main__':
    print 'Amostra de alguns números em base 28'

    l = range(11)
    l.extend([calcbase(x) for x in ('3zz', '422','4zz','522','zzz','3222')])
    l.extend([27,28,29,28**2-1,28**2,1001,1100,1200,2000])

    for n in sorted(l):
        v = reprbase(n)
        assert n == calcbase(v)
        print '%8d\t%8s' % (n, reprbase(n))




#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
Em Python os atributos de uma classe são armazenados em um `dict`, portanto
sua ordem não é preservada. Normalmente a ordem não é realmente importante.

Note no exemplo abaixo que a lista devolvida por `dir(l)` não preserva
a ordem em que foram declarados os atributos na classe `Livro`::

    >>> class LivroSimples(object):
    ...     titulo = u''
    ...     isbn = u''
    ...     autores = u''
    >>> l = LivroSimples()
    >>> dir(l) #doctest: +ELLIPSIS
    [...'autores', 'isbn', 'titulo'...]
    
Para gerar formulários automaticamente a partir da classe, é desejável
respeitar a ordem de declaração dos campos. Usando descritores e uma 
metaclasse, é possível preservar esta ordem.

    >>> class Livro(OrderedModel):
    ...     titulo = OrderedProperty()
    ...     isbn = OrderedProperty()
    ...     autores = OrderedProperty()
    >>> l2 = Livro()
    >>> l2.titulo = 'O Alienista'
    >>> l2.titulo
    'O Alienista'
    >>> list(l2)
    ['titulo', 'isbn', 'autores']
    >>> for campo in l2: print campo
    titulo
    isbn
    autores
    >>> l3 = Livro()
    >>> l3.titulo
    Traceback (most recent call last):
      ...
    AttributeError: 'Livro' object has no attribute 'titulo'
    >>> l4 = Livro(titulo=u'Alice', autores=[u'Carroll', u'Tenniel'], isbn=u'9781234567890')
    >>> for campo, valor in l4.iteritems():
    ...     print '%-8s: %s' % (campo, valor)
    titulo  : Alice
    isbn    : 9781234567890
    autores : [u'Carroll', u'Tenniel']
    
    
    
Os descritores têm um atributo `order` que é inicializado com um contador da 
classe `OrderedProperty` incrementado a cada nova instância. A metaclasse usa
este atributo `order` para ordenar uma lista com os nomes dos campos.

    >>> class Bicicleta(OrderedModel):
    ...     rodas = OrderedProperty()
    ...     aro = OrderedProperty()
    ...     cor = OrderedProperty()
    ...
    >>> bike = Bicicleta()
    >>> bike.rodas = 2
    >>> bike.aro = 26
    >>> bike.cor = u'preto'
    ...

"""
from isis.model.ordered import OrderedModel, OrderedProperty

def test():
    import doctest
    doctest.testmod()

if __name__=='__main__':
    test()


from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

requirements = ['colander', 'deform']
additional_files = ['README.txt', 'CHANGES.txt']

setup(name='isisdm', version='0.1.1',
      packages=find_packages(),
      long_description=README + "\n\n" + CHANGES,
      namespace_packages=['isis'],
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        ],
      keywords='isis application database bireme',
      author="BIREME/OPAS/OMS",
      author_email="isisnbp-devel@listas.bireme.br",
      url="http://reddes.bvsalud.org/",
      download_url='https://github.com/bireme/isisdm',
      license="LGPL v2.1 (http://www.gnu.org/licenses/lgpl-2.1.txt)",
      install_requires=requirements,
      test_suite='isis.model',
      tests_require=['Nose'],
      description='ISIS-DM allows programmers to express constraints and functionality similar to those of the CDS/ISIS Field Definition Table and the CDS/ISIS Formatting Language, but in contemporary, object-oriented programming languages.',
      data_files=additional_files,
    )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
----------------------
Object-document mapper
----------------------
    >>> def text_validator(node, value):
    ...     if value.startswith('Banana'):
    ...         raise BaseException, "You can't start a text with 'Banana'"
    ...

    >>> def colon_validator(node, value):
    ...     for author in value:
    ...         if ',' not in author:
    ...             raise BaseException, "Authors name must be in 'LastName, FirstName' format"

    >>> def is_number(node, value):
    ...     if not value.isdigit():
    ...         raise ValueError, "Value must be Integer"
    ...

    >>> class Book(Document):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = MultiTextProperty(required=False, validator=colon_validator)
    ...     pages = TextProperty()
    ...
    >>> class Article(Document):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = CompositeTextProperty(required=False, subkeys='fl')
    ...     publisher = TextProperty(required=True, validator=text_validator)
    ...
    >>> class Magazine(Document):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = MultiCompositeTextProperty(required=False, subkeys='fl')
    ...     pages = TextProperty(validator=is_number)
    ...

Using ReferenceProperty::

    >>> class Collection(Document):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     publisher = TextProperty(required=True, validator=text_validator)
    ...
    >>> class BookWithinCollection(Document):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = MultiCompositeTextProperty(required=False, subkeys='fl')
    ...     collection = ReferenceProperty()
    ...

Using class Meta::

    >>> class BookWithinMeta(Document):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = MultiTextProperty(required=False, validator=colon_validator)
    ...     pages = TextProperty()
    ...     status = TextProperty()
    ...
    ...     class Meta:
    ...         hide = ('status',)

Using FileProperty::

    >>> class BookWithAttachment(Document):
    ...     title = TextProperty(required=True, validator=text_validator)
    ...     authors = MultiCompositeTextProperty(required=False, subkeys='fl')
    ...     cover = FileProperty()
    ...

Using BooleanProperty::

    >>> class BookWithBoolean(Document):
    ...     title = TextProperty(required=True)
    ...     is_published = BooleanProperty()
    ...

Instantiating a Book object::

    >>> book1 = Book(title='Godel, Escher, Bach',
    ...               authors=(u'Hofstadter, Douglas',),
    ...               pages='777')
    ...
    >>> article1 = Article(title='Too Soon To Tell: Essays for the End of The Computer Revolution',
    ...               authors=u'^lGrier^fDavid',
    ...               publisher=u'IEEE Computer Society')
    ...
    >>> magazine1 = Magazine(title='Algorithms in a nutshell',
    ...               authors=(u'^lGreene^fLewis Joel',
    ...                        u'^lRodrigues^fJose Antunes',
    ...                        u'^lCalixto^fJoao Batista'))
    ...
    >>> book2 = BookWithinCollection(title='Ninar songs',
    ...               authors=(u'^lHeineman^fGeorge T.',
    ...                        u'^lPollice^fGary',
    ...                        u'^lSelkov^fStanley'),
    ...               collection='123abc')
    ...
    >>> book3 = BookWithAttachment(title='Other Ninar songs',
    ...               authors=(u'^lHeineman^fGeorge T.',
    ...                        u'^lPollice^fGary',
    ...                        u'^lSelkov^fStanley'),
    ...               cover={'fp':open('test_mapper.py')})
    ...
    >>> book4 = BookWithBoolean(title='Other other Ninar songs',
    ...                         is_published=True)

Manipulating its attributes::

    >>> book1.title
    u'Godel, Escher, Bach'

    >>> book1.authors[0]
    u'Hofstadter, Douglas'

    >>> book1.authors = (u'Hofstadter Douglas',)
    Traceback (most recent call last):
    ...
    BaseException: Authors name must be in 'LastName, FirstName' format

    >>> book1.authors[0]
    u'Hofstadter, Douglas'

    >>> book1.authors += (u'Daiana Rose',)
    Traceback (most recent call last):
    ...
    BaseException: Authors name must be in 'LastName, FirstName' format

    >>> book1.authors += (u'Rose, Daiana',)
    >>> book1.authors
    (u'Hofstadter, Douglas', u'Rose, Daiana')

    >>> print article1.authors
    ^lGrier^fDavid

    >>> article1.authors['f']
    u'David'

    >>> article1.authors['j']
    Traceback (most recent call last):
    ...
    KeyError: 'j'

    >>> magazine1.authors[0]['f']
    u'Lewis Joel'

    >>> magazine1.authors[0]['j']
    Traceback (most recent call last):
    ...
    KeyError: 'j'

    >>> for i in magazine1.authors: print i['l']
    Greene
    Rodrigues
    Calixto

    >>> book2.collection
    u'123abc'
    >>> book2.collection = tuple('abcdef')
    Traceback (most recent call last):
      ...
    TypeError: Reference value must be unicode or str instance

    >>> book4.is_published
    True
    >>> book4.is_published = 'must raise error'
    Traceback (most recent call last):
      ...
    TypeError: 'is_published' must be bool

Colander Schema Generation::

    >>> book1.to_python() == {'authors': (u'Hofstadter, Douglas', u'Rose, Daiana'),
    ...     'TYPE': 'Book', 'pages': u'777', 'title': u'Godel, Escher, Bach'}
    True


    >>> sc = Book.get_schema()
    >>> ' '.join('%s:%s' % (c.name, type(c.typ).__name__) for c in sc.children)
    'title:String authors:Sequence pages:String'

    >>> book_with_bool_schema = BookWithBoolean.get_schema()
    >>> ' '.join('%s:%s' % (c.name, type(c.typ).__name__) for c in book_with_bool_schema.children)
    'title:String is_published:Boolean'

Hidding an attribute

    >>> book_with_meta_schema = BookWithinMeta.get_schema()
    >>> ' '.join('%s:%s' % (c.name, type(c.typ).__name__) for c in book_with_meta_schema.children)
    'title:String authors:Sequence pages:String'
"""
from isis.model import Document
from isis.model import TextProperty, MultiTextProperty
from isis.model import CompositeTextProperty, MultiCompositeTextProperty
from isis.model import ReferenceProperty, FileProperty, BooleanProperty

def test():
    import doctest
    doctest.testmod()

if __name__=='__main__':
    test()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
-----------------
Subfield parsing
-----------------

    >>> expand('John Tenniel^rillustrator')
    [('_', 'John Tenniel'), ('r', 'illustrator')]

All subfields are stripped of trailing whitespace::

    >>> expand('John Tenniel ^rillustrator ')
    [('_', 'John Tenniel'), ('r', 'illustrator')]

Subfield keys are case-insensitive, and always converted to lower case::

    >>> expand('John Tenniel^Rillustrator')
    [('_', 'John Tenniel'), ('r', 'illustrator')]


Even if there is no main subfield, the '_' key is returned::

    >>> expand('^1one^2two^3three')
    [('_', ''), ('1', 'one'), ('2', 'two'), ('3', 'three')]

---------------
Abnormal cases
---------------

Empty field::

    >>> expand('')
    [('_', '')]

Empty subfields::

    >>> expand('aa^1^2c') # empty subfield ^1, middle position
    [('_', 'aa'), ('1', ''), ('2', 'c')]
    >>> expand('aa^1b^2') # empty subfield ^2, last position
    [('_', 'aa'), ('1', 'b'), ('2', '')]

Subfield keys are limited to a-z and 0-9. Invalid keys are ignored,
and the subfield delimiter is appended to the previous subfield::

    >>> expand('John Tenniel^!illustrator')
    [('_', 'John Tenniel^!illustrator')]
    >>> expand('John Tenniel^rillustrator^')
    [('_', 'John Tenniel'), ('r', 'illustrator^')]
    >>> expand('John Tenniel^rillustrator^^')
    [('_', 'John Tenniel'), ('r', 'illustrator^^')]

To reduce the damage from duplicate delimiters in the middle of the
field, a space is added after each pair. Otherwise the example below
would seem to have an ^i subfield with the content "llustrator".
Keeping the duplicate delimiters together makes it easier to find
and fix these problems later::

    >>> expand('John Tenniel^^illustrator')
    [('_', 'John Tenniel^^ illustrator')]

----------------------------
Controlled subfield parsing
----------------------------

When a subkeys parameter is passed, only subfield markers with those keys
are expanded::

    >>> expand('John Tenniel^rillustrator', subkeys='r')
    [('_', 'John Tenniel'), ('r', 'illustrator')]
    >>> expand('John Tenniel^xillustrator', subkeys='r')
    [('_', 'John Tenniel^xillustrator')]
    >>> expand('John Tenniel^rillustrator', subkeys='')
    [('_', 'John Tenniel^rillustrator')]


---------------------
CompositeString tests
---------------------
    >>> author = CompositeString('John Tenniel^rillustrator',
    ...                          subkeys='r')
    
    >>> unicode(author)
    u'John Tenniel^rillustrator'
    
    >>> str(author)
    'John Tenniel^rillustrator'
    
Iteration semantics are similar to dicts::

    >>> for k in author: print k
    _
    r
    
    >>> for k, v in author.items(): print k, v
    _ John Tenniel
    r illustrator
"""

from isis.model.subfield import expand, CompositeString
import json

def test():
    import doctest
    doctest.testmod()

if __name__=='__main__':
    test()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# ISIS-DM: the ISIS Data Model API
#
# Copyright (C) 2010 BIREME/PAHO/WHO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from operator import attrgetter

class OrderedProperty(object):
    __count = 0

    def __init__(self):
        self.order = OrderedProperty.__count
        OrderedProperty.__count += 1

    def __get__(self, instance, cls):
        try:
            return instance._prop_values[self.name]
        except KeyError:
            raise AttributeError("'%s' object has no attribute '%s'" % (instance.__class__.__name__, self.name))

    def __set__(self, instance, value):
        instance._prop_values[self.name] = value

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

class OrderedMeta(type):
    def __new__(cls, name, bases, dict):
        if 'TYPE' in dict:
            raise TypeError('TYPE is a reserved identifier.')
        dict['TYPE'] = name
        props = []
        for key, attr in dict.items():
            if isinstance(attr, OrderedProperty):
                attr.name = key
                props.append(attr)
        dict['_ordered_props'] = sorted(props, key=attrgetter('order'))
        return type.__new__(cls, name, bases, dict)

    def __iter__(self):
        return (prop.name for prop in self._ordered_props)

    def __contains__(self, key):
        return key in self._ordered_props

class OrderedModel(object):
    __metaclass__ = OrderedMeta

    def __init__(self, **kwargs):
        self._prop_values = {}
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __iter__(self):
        return iter(self.__class__)

    def iteritems(self):
        return ((key, getattr(self, key))
                for key in self.__class__)


if __name__=='__main__':
    import doctest
    doctest.testmod()
